@@ -72,8 +72,7 @@ class BaseLineParser(QObject):
         Return:
             True if the file should be saved, False otherwise.
         """
-        if not os.path.exists(self._configdir):
-            os.makedirs(self._configdir, 0o755)
+        os.makedirs(self._configdir, 0o755, exist_ok=True)
         return True
 
     def _after_save(self):
